﻿using System;
using System.Collections.Generic;
using System.Linq;
using CardPlay.Model.Primitives;

namespace CardPlay.Model.Layout
{
    public class GridLayout : GridLayoutBase
    {
        private readonly Dictionary<int, double> _horzScalingFactors = new Dictionary<int, double>();
        private readonly Dictionary<int, double> _vertScalingFactors = new Dictionary<int, double>();

        private int[] _columnWidths;
        private int[] _rowHeights;

        public double? GetColumnScalingFactor(int column)
        {
            return _horzScalingFactors.ContainsKey(column) ? (double?) _horzScalingFactors[column] : null;
        }

        public void SetColumnScalingFactors(params double?[] list)
        {
            for (var i = 0; i < Math.Min(list.Length, Columns); ++i)
                SetColumnScalingFactor(i, list[i]);
        }

        public void SetColumnScalingFactor(int column, double? scalingFactor)
        {
            if (scalingFactor.HasValue)
                _horzScalingFactors[column] = scalingFactor.Value;
            else
                _horzScalingFactors.Remove(column);
        }

        public double? GetRowScalingFactor(int column)
        {
            return _vertScalingFactors.ContainsKey(column) ? (double?)_vertScalingFactors[column] : null;
        }

        public void SetRowScalingFactors(params double?[] list)
        {
            for (var i = 0; i < Math.Min(list.Length, Columns); ++i)
                SetRowScalingFactor(i, list[i]);
        }
        public void SetRowScalingFactor(int row, double? scalingFactor)
        {
            if (scalingFactor.HasValue)
                _vertScalingFactors[row] = scalingFactor.Value;
            else
                _vertScalingFactors.Remove(row);
        }

        protected override Size OnMeasure(ILayoutContext context, Size availableSize)
        {
            var size = base.OnMeasure(context, availableSize);
            if (Columns == 0 || Rows == 0) return size;

            foreach (var visibleChild in VisibleChildren)
                visibleChild.Measure(context, availableSize);

            _columnWidths = HorizontalAxisLayout;
            _rowHeights = VerticalAxisLayout;

            ApplyScaling(availableSize);

            return new Size(_columnWidths.Sum(), _rowHeights.Sum());
        }
        
        protected override void OnArrange(ILayoutContext context, Rectangle bounds)
        {
            if (Columns == 0 || Rows == 0) return;

            var contentSize = new Size(_columnWidths.Sum(), _rowHeights.Sum());
            var origin = contentSize.PositionWithin(bounds, HorizontalContentAlignment, VerticalContentAlignment).Location;

            var col = 0;
            var row = 0;

            foreach (var layout in Children)
            {
                if (!layout.Collapsed)
                {
                    var cellBounds = CalcCellBounds(origin, col, row);
                    layout.Arrange(context, cellBounds);
                }

                if (++col == Columns)
                {
                    col = 0;
                    row++;
                }
            }
        }

        private void ApplyScaling(Size bounds)
        {
            var size = new Size(_columnWidths.Sum(), _rowHeights.Sum());

            if (bounds.Width != size.Width)
            {
                var variableSize = _horzScalingFactors.Keys.Select(n => _columnWidths[n]).Sum();

                if (variableSize > 0)
                {
                    var weight = _horzScalingFactors.Values.Sum();
                    variableSize = Math.Max(0, bounds.Width - (size.Width - variableSize));

                    foreach (var item in _horzScalingFactors)
                        _columnWidths[item.Key] = (int)(variableSize * item.Value / weight);
                }
            }

            if (bounds.Height != size.Height)
            {
                var variableSize = _vertScalingFactors.Keys.Select(n => _rowHeights[n]).Sum();

                if (variableSize > 0)
                {
                    var weight = _vertScalingFactors.Values.Sum();
                    variableSize = Math.Max(0, bounds.Height - (size.Height - variableSize));

                    foreach (var item in _vertScalingFactors)
                        _rowHeights[item.Key] = (int)(variableSize * item.Value / weight);
                }
            }
        }

        private Rectangle CalcCellBounds(Point origin, int col, int row)
        {
            return _columnWidths == null || _rowHeights == null
                ? new Rectangle(origin, Size.Empty)
                : new Rectangle(origin.X + _columnWidths.Take(col).Sum(), origin.Y + _rowHeights.Take(row).Sum(), _columnWidths[col], _rowHeights[row]);
        }
    }
}
