﻿using System;
using System.Linq;
using CardPlay.Model.Primitives;

namespace CardPlay.Model.Layout
{
    public class SideBySideLayout : CompositeLayout
    {
        private int _minimumSpacing;
        private int _maximumSpacing = 10; 

        public SpreadDirection Direction {get;set;}
    
        public int MinimumSpacing
        {
            get { return _minimumSpacing; }
            set 
            { 
                Guard.Against<ArgumentOutOfRangeException>(value > MaximumSpacing, "value");
                _minimumSpacing = value; 
            }
        }

        public int MaximumSpacing
        {
            get { return _maximumSpacing; }
            set 
            { 
                Guard.Against<ArgumentOutOfRangeException>(value < MinimumSpacing, "value");
                _maximumSpacing = value; 
            }
        }

        protected override Size OnMeasure(ILayoutContext context, Size availableSize)
        {
            if (Collapsed || !Children.Any()) return Size.Empty;

            var gaps = VisibleChildren.Count() - 1;
            var availableSlotSize = availableSize;

            var size = Size.Empty;
            var maxSize = Size.Empty;

            switch (Direction)
            {
                case SpreadDirection.Horizontal:
                {
                    availableSlotSize.Width = Math.Max(0, availableSlotSize.Width - gaps * MinimumSpacing);

                    foreach (var layout in VisibleChildren)
                    {
                        layout.Measure(context, availableSlotSize);
                        size.Width += layout.MeasuredSize.Width;
                        maxSize.Height = Math.Max(maxSize.Height, layout.MeasuredSize.Height);
                    }

                    maxSize.Width = size.Width + gaps * MaximumSpacing;
                    size = new Size(size.Width + gaps * MinimumSpacing, maxSize.Height);
                    break;
                }

                case SpreadDirection.Vertical:
                {
                    availableSlotSize.Height = Math.Max(0, availableSlotSize.Height - gaps * MinimumSpacing);

                    foreach (var layout in VisibleChildren)
                    {
                        layout.Measure(context, availableSlotSize);
                        size.Height += layout.MeasuredSize.Height;
                        maxSize.Width = Math.Max(maxSize.Width, layout.MeasuredSize.Width);
                    }

                    maxSize.Height = size.Height + gaps * MaximumSpacing;
                    size = new Size(maxSize.Width, size.Height + gaps * MinimumSpacing);
                    break;
                }
            }

            return availableSize.ConstrainedTo(size, maxSize);
        }

        protected override void OnArrange(ILayoutContext context, Rectangle bounds)
        {
            if (! Children.Any()) return;

            var contentSize = VisibleChildren.Aggregate(Size.Empty, (result, site) => result + site.MeasuredSize);
            var gaps = VisibleChildren.Count() - 1;

            switch (Direction)
            {
                case SpreadDirection.Horizontal:
                {
                    var spacing = gaps > 0 ? Math.Min(Math.Max(MinimumSpacing, (bounds.Width - contentSize.Width) / gaps), MaximumSpacing) : 0;

                    contentSize.Width += spacing * gaps;
                    contentSize.Height = Bounds.Height;

                    var origin = contentSize.PositionWithin(bounds, HorizontalContentAlignment, VerticalContentAlignment).Location;

                    foreach (var layout in VisibleChildren)
                    {
                        var width = layout.MeasuredSize.Width;

                        layout.Arrange(context, new Rectangle(origin.X, origin.Y, width, Bounds.Height));
                        origin.X += width + spacing;
                    }                
                    break;
                }
                
                case SpreadDirection.Vertical:
                {
                    var spacing = gaps > 0 ? Math.Min(Math.Max(MinimumSpacing, (Bounds.Height - contentSize.Height) / gaps), MaximumSpacing) : 0;

                    contentSize.Width = Bounds.Width;
                    contentSize.Height += spacing * gaps;

                    var origin = contentSize.PositionWithin(Bounds, HorizontalContentAlignment, VerticalContentAlignment).Location;

                    foreach (var layout in VisibleChildren)
                    {
                        var height = layout.MeasuredSize.Height;

                        layout.Arrange(context, new Rectangle(origin.X, origin.Y, Bounds.Width, height));
                        origin.Y += height + spacing;
                    }
                    break;
                }
            }
        }
    }
}
